open Core
open Syslog
open Okweb

let from_env ~conv ~key ~default =
  try
    Sys.getenv key |> Option.map ~f:conv |> Option.value ~default
  with
  | _ ->
    sprintf "Warning: invalid env variable: %s. Using default value." key |> syslog;
    default

let from_env_or_abort ~conv ~key =
  let value_opt = Sys.getenv key in
  try
    Option.value_exn (
      value_opt
      |> Option.filter ~f:(fun s -> not (String.is_empty s))
      |> Option.map ~f:conv)
  with
  | _ ->
    sprintf "Error: invalid or absent env variable: %s. Exiting." key |> syslog;
    exit 1

let get_config () =
  let open Okauth_crypto in
  let open Okauth_client in
  let base_url = from_env_or_abort ~key:"BASE_URL" ~conv:Fn.id in
  let app_name = "OkDiary" in
  let public_encryption_key =
    from_env_or_abort
      ~key:"PUBLIC_ENCRYPTION_KEY"
      ~conv:Public_key.of_base64_exn in
  let secret_encryption_key =
    from_env_or_abort
      ~key:"SECRET_ENCRYPTION_KEY"
      ~conv:Secret_key.of_base64_exn in
  let public_signing_key =
    from_env_or_abort
      ~key:"PUBLIC_SIGNING_KEY"
      ~conv:Public_key.of_base64_exn in
  let secret_signing_key =
    from_env_or_abort
      ~key:"SECRET_SIGNING_KEY"
      ~conv:Secret_key.of_base64_exn in
  let storage = In_memory_storage.empty in
  { base_url;
    app_name;
    public_encryption_key;
    secret_encryption_key;
    public_signing_key;
    secret_signing_key;
    storage }

let serve () =
  openlog ~id:"okdiaryd" ~options:[ Open_option.PID; Open_option.PERROR; ] ();
  let address, port, db_path = (
    from_env ~key:"HOST" ~default:"localhost" ~conv:Fn.id,
    from_env ~key:"PORT" ~default:4201 ~conv:Int.of_string,
    from_env ~key:"DB_PATH" ~default:"okdiary.db" ~conv:Fn.id
  ) in
  Oksql.with_db db_path begin fun db ->
    Oksql.Migration.apply db ~migrations:Data.Schema.migrations;
    let module Entry_repo = Data.Entry_repo.Make(struct let db = db end) in
    let module Auth = Okauth_client.Make(Okweb.Client)(Okauth_client.In_memory_storage) in
    let module Daemon = Okdiary.Daemon.Make(Entry_repo)(Okauth_client.In_memory_storage)(Auth) in
    let auth_config = get_config () in
    sprintf "Starting OkDiary Daemon v%s on %s:%d" [%blob "../../version"] address port |> syslog;
    Lwt_main.run (Server.listen ~address ~port (Daemon.process_request auth_config));
  end

let generate_application_keys () =
  let open Okauth_crypto in
  let (aepk, aesk) = generate_keypair ~purpose:Encryption in
  let (aspk, assk) = generate_keypair ~purpose:Signing in
  printf "PUBLIC_ENCRYPTION_KEY=\"%s\"\n" (Public_key.to_base64 aepk);
  printf "SECRET_ENCRYPTION_KEY=\"%s\"\n" (Secret_key.to_base64 aesk);
  printf "PUBLIC_SIGNING_KEY=\"%s\"\n" (Public_key.to_base64 aspk);
  printf "SECRET_SIGNING_KEY=\"%s\"\n" (Secret_key.to_base64 assk)

let () = match Sys.get_argv () with
  | [| _; "serve" |] ->
    serve ()
  | [| _; "generate-application-keys" |] ->
    generate_application_keys ()
  | [| _ |] ->
    Caml.prerr_endline "No command given. See 'man 8 okdiaryd' for available commands"
  | argv ->
    let drop_fst l = List.drop l 1 in
    let args = argv |> Array.to_list |> drop_fst |> String.concat ~sep:" " in
    eprintf "Invalid arguments: %s. See 'man 8 okdiaryd' for available commands and arguments\n" args
